# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import click
import logging
import os
import pkg_resources
import sys
import yaml

from cube import spec
from cube.client import api
from cube.daemon.api import app, db
from cube.daemon.config import ApplicationConfig
from cube.daemon.utils.pidfile import Pidfile

from nr.databind.core import ObjectMapper
from nr.databind.json import JsonModule
from nr.utils.process import replace_stdio, spawn_daemon

logger = logging.getLogger(__name__)
mapper = ObjectMapper(JsonModule())
DEFAULT_CONFIG_FILENAME = 'var/conf/runtime.yml'


def _get_config_template() -> str:
  return pkg_resources.resource_string(__name__, 'config-template.yml').decode('utf8')


def _load_config(dev: bool = False) -> ApplicationConfig:
  if dev and not os.path.isfile(DEFAULT_CONFIG_FILENAME):
    logger.info('Falling back to default configuration due to --dev')
    config_file = None
    config_data = _get_config_template()
  else:
    config_file = DEFAULT_CONFIG_FILENAME
    with open(config_file or DEFAULT_CONFIG_FILENAME) as fp:
      config_data = fp.read()

  config = ApplicationConfig.load(mapper, config_file, config_data)
  config = config.canonical(os.getcwd())
  return config


@click.group()
def cli():
  " Entrypoint for cube-daemon."

  logging.basicConfig(
    format='[%(asctime)s - %(name)s - %(levelname)s]: %(message)s',
    level=logging.INFO)


@cli.command()
@click.option('-f', '--force', is_flag=True)
def init(force):
  " Bootstrap a Cube-Daemon config in the current directory. "

  config_fn = DEFAULT_CONFIG_FILENAME
  if not os.path.isfile(config_fn) or force:
    print('Write', config_fn)
    os.makedirs(os.path.dirname(config_fn), exist_ok=True)
    with open(config_fn, 'w') as fp:
      fp.write(_get_config_template())
  else:
    print('Skip', config_fn)


@cli.command()
@click.option('--dev', is_flag=True, help='Enables dev mode (config_file argument is optional). This enables the application debug mode and reloader.')
@click.option('--detach', is_flag=True, help='Detach the process from the terminal.')
def start(dev, detach):
  " Start the service. "

  config = _load_config(dev)
  db.initialize(config.database)

  app.config['DEBUG'] = dev
  app.config['CUBE_DAEMON_SPEC_CONTEXT'] = spec.Context(
    mapper,
    products_install_directory=config.deploy.products_install_directory,
    services_install_directory=config.deploy.services_install_directory)
  app.config['CUBE_DAEMON_CONFIG'] = config
  app.config['CUBE_DAEMON_OBJECT_MAPPER'] = mapper

  pidfile = Pidfile(config.deploy.pidfile, for_realz=not dev)
  if pidfile.running():
    logger.error(
      'Another instance of cube-daemon is already running (pid: %s)',
      pidfile.get())
    sys.exit(1)

  if not dev:
    os.makedirs(os.path.dirname(config.deploy.stdout), exist_ok=True)
    os.makedirs(os.path.dirname(config.deploy.stderr), exist_ok=True)
    stdout = open(config.deploy.stdout, 'a+')
    stderr = open(config.deploy.stderr, 'a+')

  def run():
    if not dev:
      replace_stdio(None, stdout, stderr)
    with pidfile:
      app.run(host=config.server.host, port=config.server.port, debug=dev)

  if detach:
    spawn_daemon(run)
  else:
    run()


@cli.command()
def status():
  config = _load_config()
  pidfile = Pidfile(config.deploy.pidfile)
  if pidfile.running():
    print('Running (pid: {})'.format(pidfile.get()))
  else:
    print('Stopped')


@cli.command()
def stop():
  config = _load_config()
  pidfile = Pidfile(config.deploy.pidfile)
  if not pidfile.running():
    logger.error('There is no running instance of cube-daemon.')
    sys.exit(1)
  pidfile.terminate()


@cli.command()
@click.option('--product-path', default='.', help='The path to the product to test.')
@click.option('--service-home', default=None, help='The home directory of the service. Defaults to ".<product-path-base>-home".')
@click.option('--config-overrides', default=None, help='The config overrides. Defaults to "config-overrides.yml" in the service home directory.')
@click.argument('command')
def develop(product_path, service_home, config_overrides, command):
  """
  Run a single instance of a product without the cube-daemon server. This is
  useful when developing Cube-compatible services to replicate the conditions
  under which a service would be executed.
  """

  product_path = os.path.abspath(os.path.normpath(product_path))
  if not service_home:
    service_home = '.' + os.path.basename(product_path) + '-home'
  service_home = os.path.abspath(os.path.normpath(service_home))
  if not config_overrides:
    config_overrides = os.path.join(service_home, 'config-overrides.yml')

  context = spec.Context.dev(mapper)
  product = context.product(os.path.basename(product_path), '1.0.0-dev', product_path)
  service = context.service(os.path.basename(service_home), service_home, product)
  product.validate_layout()

  do_reload = False
  if not os.path.isfile(service.runtime_yml):
    do_reload = True
  elif os.path.isfile(config_overrides) and os.path.getmtime(config_overrides) > os.path.getmtime(service.runtime_yml):
    do_reload = True
  elif os.path.getmtime(product.config_yml) > os.path.getmtime(service.runtime_yml):
    do_reload = True

  if not os.path.exists(service_home):
    service.product = None
    service.install_product(product)
    service.product = product

  if command == 'reload' or (do_reload and command != 'stop'):
    if os.path.isfile(config_overrides):
      with open(config_overrides) as fp:
        overrides = mapper.deserialize(yaml.safe_load(fp), api.ConfigOverride)
    else:
      overrides = api.ConfigOverride()
    runtime_config = spec.generate_runtime_config(
      service.name, product.get_config(), overrides, spec.NoRoleDiscovery())
    service.update_config(runtime_config)
    if command == 'reload':
      return

  if command == 'start':
    service.start()
  elif command == 'stop':
    service.stop()
  elif command == 'restart':
    service.stop()
    service.start()
  elif command == 'status':
    print(service.get_status())
  else:
    logger.error('unknown command: {!r}'.format(command))
    sys.exit(1)


if __name__ == '__main__':
  cli()
