# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import nr.fs
import os
import re
import requests
import shutil

from .common import validate_identifier, MUST_EXIST, MUST_NOT_EXIST, MAY_EXIST
from cube.client import api
from cube.daemon.api import spec_context, db, mapper
from cube.daemon.utils.hashing import hash_file
from cube.daemon.utils.logging import class_logger
from cube.spec import DirectoryInstallMode
from nr.databind.rest import InvalidArgument
from nr.interface import implements
from nr.utils.ponyorm.utils import upsert


def get_product(name: str, version: str, mode: str = MUST_EXIST) -> db.Product:
  product = db.Product.get(name=name, version=version)
  if not product and mode == MUST_EXIST:
    raise api.ResourceDoesNotExist(type='product',
      productName=name, productVersion=version)
  elif product and mode == MUST_NOT_EXIST:
    raise api.ResourceAlreadyExists(type='product',
      productName=name, productVersion=version)
  return product


@implements(api.ProductsService)
class ProductsServiceImpl:

  logger = class_logger()

  @db.session
  def get_products(self):
    return list(map(lambda p: p.to_api(), db.Product.select()))

  @db.session
  def get_product_versions(self, name):
    return list(map(lambda p: p.to_api(),
      db.Product.select(lambda p: p.name == name)))

  @db.session
  def get_product(self, name, version):
    return get_product(name, version).to_api()

  @db.session
  def delete_product(self, name, version):
    product = get_product(name, version)
    services = list(db.select(s.name for s in product.services))
    if services:
      raise api.ResourceBusy(
        type='product', productName=name, productVersion=version,
        usedByServices=services)
    product.to_spec().delete()
    product.delete()

  @db.session
  def import_product_from_url(self, name, version, tar_url, force):
    response = requests.get(tar_url, stream=True)
    match = re.search(r'filename=([^;]+)', response.headers.get('Content-Disposition', ''))
    if match:
      filename = match.group(1)
    else:
      filename = 'unknown-filename.tgz'
    self.logger.info('Downloaded file "%s" from "%s".', filename, tar_url)
    with nr.fs.tempfile() as tempfile:
      shutil.copyfileobj(response.raw, tempfile)
      tempfile.close()
      with open(tempfile.name, 'rb') as fp:
        fp.filename = filename
        return self.import_product_from_tar(name, version, fp, force, _origin=tar_url)

  @db.session(optimistic=False)
  def import_product_from_tar(self, name, version, tar, force, _origin: str = None):
    validate_identifier('name', name)
    validate_identifier('version', version)
    if not force:
      get_product(name, version, MUST_NOT_EXIST)
    else:
      product = get_product(name, version, MAY_EXIST)
      if product:
        product.to_spec().delete()

    with nr.fs.tempfile() as tempfile:
      shutil.copyfileobj(tar, tempfile)
      tempfile.close()
      sha1sum = hash_file(tempfile.name, 'sha1')
      md5sum = hash_file(tempfile.name, 'md5')
      spec_product = spec_context.product(name, version)
      spec_product.install_from_tar(tempfile.name)

    product = upsert(db.Product, {'name': name, 'version': version}, {
      'path': spec_product.path,
      'status': api.ProductStatus.INSTALLED,
      'origin': _origin or 'tar://{},sha1:{},md5:{}'.format(tar.filename, sha1sum, md5sum),
      'config': mapper.serialize(spec_product.get_config(), api.ProductConfig),
      'cache_config': True
    })[0]

    return product.to_api()

  @db.session(optimistic=False)
  def import_product_from_path(self, name, version, path, symlink, force):
    validate_identifier('name', name)
    validate_identifier('version', version)
    if not force:
      get_product(name, version, MUST_NOT_EXIST)
    else:
      product = get_product(name, version, MAY_EXIST)
      if product:
        product.to_spec().delete()

    if not os.path.isabs(path):
      raise InvalidArgument('path must be an absolute path on the host.',
        parameters={'path': path})

    if not os.path.isdir(path):
      raise InvalidArgument('specified path does not exist on the host.',
        parameters={'path': path})

    if symlink:
      # Check if we don't already have this path saved as another product.
      # We don't allow two products pointing to the same path.
      product = db.Product.get(path=path)
      if product:
        raise InvalidArgument('path is already used by another product',
          parameters={'path': path, 'productName': product.name, 'productVersion': product.version})

    install_mode = DirectoryInstallMode.Symlink if symlink else \
      DirectoryInstallMode.Copy
    spec_product = spec_context.product(name, version)
    spec_product.install_from_directory(path, install_mode)

    if symlink:
      config = {}
    else:
      config = mapper.serialize(spec_product.get_config(), api.ProductConfig)

    product = upsert(db.Product, {'name': name, 'version': version}, {
      'path': spec_product.path,
      'status': api.ProductStatus.INSTALLED,
      'origin': 'file://' + path,
      'config': config,
      'cache_config': not symlink
    })[0]

    return product.to_api()
