# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import os

from .common import validate_identifier, MUST_EXIST, MUST_NOT_EXIST
from .products import get_product
from cube.client import api
from cube.daemon.api import spec_context, db
from cube.daemon.utils.logging import class_logger
from nr.interface import implements


def get_service(name: str, mode: str = MUST_EXIST) -> db.Service:
  service = db.Service.get(name=name)
  if not service and mode == MUST_EXIST:
    raise api.ResourceDoesNotExist(type='service', serviceName=name)
  elif service and mode == MUST_NOT_EXIST:
    raise api.ResourceAlreadyExists(type='service', serviceName=name)
  return service


@implements(api.ServicesService)
class ServicesServiceImpl:
  logger = class_logger()

  @db.session
  def get_services(self):
    return list(map(lambda s: s.to_api(), db.Service.select()))

  @db.session
  def get_service(self, name):
    return get_service(name).to_api()

  @db.session
  def create_service(self, name):
    validate_identifier('name', name)
    get_service(name, MUST_NOT_EXIST)

    spec_service = spec_context.service(name)
    if os.path.isdir(spec_service.home):
      raise RuntimeError('service home directory {!r} already exists'.format(
        spec_service.home))

    service = db.Service(name=name, home=spec_service.home)
    return service.to_api()

  @db.session
  def delete_service(self, name):
    service = get_service(name)
    if service.status not in (api.ServiceStatus.CREATED, api.ServiceStatus.STOPPED):
      raise api.IllegalState(serviceName=name, status=service.status.name)
    service.to_spec().delete()
    service.delete()

  def install_product(self, name, product):
    with db.session:
      service = get_service(name)
      product = get_product(product.name, product.version)
      spec_service = service.to_spec()
      status = service.update_status(spec_service)
      if status == api.ServiceStatus.UNKNOWN:
        raise api.IllegalState(serviceName=name, status=status.name)

    # TODO (@NiklasRosenstein): Move to a Plan/Task

    if status == api.ServiceStatus.RUNNING:
      spec_service.stop()
      with db.session:
        service = get_service(name)
        service.update_status(spec_service)
      assert service.status == api.ServiceStatus.STOPPED

    spec_service.install_product(product.to_spec())

    with db.session:
      service = get_service(name)
      service.installed_product = get_product(product.name, product.version)
      service.update_status(spec_service)
      runtime_config = service.get_runtime_config()

    self.logger.info('Updating runtime configuration of service %r', service.name)
    spec_service.update_config(runtime_config)

    if status == api.ServiceStatus.RUNNING:
      spec_service.start()

    with db.session:
      service = get_service(name)
      return service.to_api()

  # TODO (@NiklasRosenstein): Getting error if optimistic=True (default):
  #     OptimisticCheckError: Object Service['xyz'] was updated outside of current transaction
  #   Probably because of the orm.commit() call in Service.update_status()
  @db.session(optimistic=False)
  def manage_lifecycle(self, name, stop, start):
    service = get_service(name)
    if not service.installed_product:
      raise api.IllegalState('Service has no product installed.', serviceName=name)
    spec_service = service.to_spec()
    status = service.update_status(spec_service)
    if stop and status == api.ServiceStatus.RUNNING:
      spec_service.stop()
      status = service.update_status(spec_service)
    if start and status != api.ServiceStatus.RUNNING:
      spec_service.start()
      status = service.update_status(spec_service)
    return service.to_api()

  @db.session
  def get_plans(self, name, page_start):
    raise NotImplementedError

  @db.session
  def get_latest_plan(self, name):
    raise NotImplementedError
