# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import copy
import os
import yaml

from nr.databind.core import Field, FieldName, Struct, ObjectMapper
from nr.databind.core.decorations import Remainder, Validator
from nr.databind.json import JsonModule
from typing import Optional


def _canonical(path: Optional[str], parent: str) -> Optional[str]:
  if path is not None:
    path = os.path.abspath(os.path.normpath(os.path.join(parent, path)))
  return path


class ServerConfig(Struct):
  host = Field(str, default='localhost')
  port = Field(int, default=6123)
  debug = Field(bool, default=False)


class DatabaseConfig(Struct):
  provider = Field(str)
  filename = Field(str, default=None)
  options = Field(dict, Remainder())

  def canonical(self, parent_directory: str) -> 'DatabaseConfig':
    self = copy.copy(self)
    self.filename = _canonical(self.filename, parent_directory)
    return self

  def bind_kwargs(self) -> dict:
    return {'provider': self.provider, 'filename': self.filename, **self.options}


class DeployConfig(Struct):
  products_install_directory = Field(str, FieldName('products-install-directory'), default='var/products')
  services_install_directory = Field(str, FieldName('services-install-directory'), default='var/services')
  tasks_data_directory = Field(str, FieldName('tasks-data-directory'), default='var/data/tasks')
  pidfile = Field(str, default='var/run/cube-daemon.pid')
  stdout = Field(str, default='var/log/stdout.log')
  stderr = Field(str, default='var/log/stderr.log')

  def canonical(self, parent_directory: str) -> 'DeployConfig':
    self = copy.deepcopy(self)
    self.products_install_directory = _canonical(self.products_install_directory, parent_directory)
    self.services_install_directory = _canonical(self.services_install_directory, parent_directory)
    self.tasks_data_directory = _canonical(self.tasks_data_directory, parent_directory)
    self.pidfile = _canonical(self.pidfile, parent_directory)
    self.stdout = _canonical(self.stdout, parent_directory)
    self.stderr = _canonical(self.stderr, parent_directory)
    return self


class ApplicationConfig(Struct):
  database = Field(DatabaseConfig)
  deploy = Field(DeployConfig, default=Field.DEFAULT_CONSTRUCT)
  server = Field(ServerConfig)

  def canonical(self, parent_directory: str) -> 'ApplicationConfig':
    return ApplicationConfig(self.database.canonical(parent_directory),
      self.deploy.canonical(parent_directory), self.server)

  @classmethod
  def load(cls, mapper: ObjectMapper, filename: str, data: str = None) -> 'ApplicationConfig':
    """
    Loads the application config from a file. The config is expected to be
    nested in a "$.runtime" key.
    """

    if data is None:
      with open(filename) as fp:
        data = fp.read()
    data = yaml.safe_load(data)
    return mapper.deserialize(data, {'runtime': Field(cls)}, filename=filename).runtime
