# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
from .logging import class_logger
from nr.utils.process import process_exists, process_terminate
from typing import Optional


class Pidfile:
  """
  A helper class to manage PID files. If *for_realz* is set to False, it will
  not actually save data and act as if the PID file does not exist even if it
  does.
  """

  logger = class_logger()

  def __init__(self, filename: str, create_dir: bool = True, for_realz: bool = True):
    self.filename = filename
    self.create_dir = create_dir
    self.for_realz = for_realz
    self._cache = None

  def __enter__(self):
    self.save()

  def __exit__(self, *args):
    self.rm()

  def save(self, pid: int = None):
    if not self.for_realz:
      return
    if pid is None:
      pid = os.getpid()
    if self.create_dir:
      os.makedirs(os.path.dirname(self.filename or '.'), exist_ok=True)
    with open(self.filename, 'w') as fp:
      fp.write(str(pid))
    self._cache = pid

  def get(self) -> Optional[int]:
    if not self.for_realz:
      return None
    if self._cache is None and os.path.isfile(self.filename):
      with open(self.filename, 'r') as fp:
        data = fp.readline()
      try:
        self._cache = int(data)
      except ValueError as exc:
        self.logger.error('Unable to read process ID from "%s".', self.filename)
    return self._cache

  def rm(self):
    if self.for_realz and os.path.isfile(self.filename):
      try:
        os.remove(self.filename)
      except OSError as exc:
        self.logger.exception('Unable to remove pidfile "%s".', self.filename)

  def terminate(self, timeout: int = 10):
    pid = self.get()
    if pid is not None:
      process_terminate(pid, timeout=timeout)

  def running(self) -> bool:
    if not self.for_realz:
      return False
    pid = self.get()
    if pid is not None:
      return process_exists(pid)
    return False
